{**************************************************************************************************}
{                                                                                                  }
{  TimeZones - win control paint box                                                            }
{                                                                                                  }
{  The contents of this file are subject to the Y Library Public License Version 1.0 (the          }
{  "License"); you may not use this file except in compliance with the License. You may obtain a   }
{  copy of the License at http://delphi.pjh2.de/                                                   }
{                                                                                                  }
{  Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF  }
{  ANY KIND, either express or implied. See the License for the specific language governing        }
{  rights and limitations under the License.                                                       }
{                                                                                                  }
{  The Original Code is: TimeZones.pas.                                                            }
{  The Initial Developer of the Original Code is Peter J. Haas (libs@pjh2.de). Portions created    }
{  by Peter J. Haas are Copyright (C) 2000-2005 Peter J. Haas. All Rights Reserved.                }
{                                                                                                  }
{  Contributor(s):                                                                                 }
{                                                                                                  }
{  You may retrieve the latest version of this file at the homepage of Peter J. Haas, located at   }
{  http://delphi.pjh2.de/                                                                          }
{                                                                                                  }
{**************************************************************************************************}

// For history see end of file

{$ALIGN ON, $BOOLEVAL OFF, $LONGSTRINGS ON, $IOCHECKS ON, $WRITEABLECONST OFF}
{$OVERFLOWCHECKS OFF, $RANGECHECKS OFF, $TYPEDADDRESS ON, $MINENUMSIZE 1}

unit TimeZones;

// Tested with:
//
// - Delphi 3.02 Professional
// - Win98 Second Edition

interface
uses
  Windows, SysUtils, Classes;

const
  TIME_ZONE_ID_INVALID  = DWORD($FFFFFFFF);
  TIME_ZONE_ID_UNKNOWN  = 0;
  TIME_ZONE_ID_STANDARD = 1;
  TIME_ZONE_ID_DAYLIGHT = 2;
  
type
  TTimeZonesList = class(TObject)
  private
    FList : TStringList;
    FCurrent : Integer;
    function GetCount:Integer;
    function GetItems(Index:Integer):TTimeZoneInformation;
    procedure ClearList;
    procedure ReadList;
  public
    constructor Create;
    destructor Destroy; override;
    function IndexOf(const AName:String):Integer;
    procedure AssignList(AList:TStrings);
    property Count:Integer read GetCount;
    property Items[Index:Integer]:TTimeZoneInformation read GetItems;
    property Current:Integer read FCurrent;
  end;

function TimeZoneBiasToStr(ABias:SmallInt):String;
function TimeZoneDateToStr(ADate:TSystemTime):String;

function TimeZoneDateToDateTime(ADate:TSystemTime; AYear:Word):TDateTime;
function TimeZoneDateToDateTimeStr(ADate:TSystemTime; AYear:Word):String;

implementation
uses
  Registry;

{$ifdef Ver90}
const
{$else}
resourcestring
{$endif}
  sTZFirst  = 'Erster';
  sTZSecond = 'Zweiter';
  sTZThird  = 'Dritter';
  sTZFourth = 'Vierter';
  sTZLast   = 'Letzter';
  sTZDateDesc = '%s %s im %s um %s';
  sTZNotSet   = 'nicht gesetzt fr diese Zeitzone';

function TimeZoneBiasToStr(ABias:SmallInt):String;
begin
  if ABias>0 then begin
    Result:=FormatDateTime('"-"hh : nn',ABias/(24*60));
  end
  else begin
    Result:=FormatDateTime('"+"hh : nn',-ABias/(24*60));
  end;
end;

function TimeZoneDateToStr(ADate:TSystemTime):String;
const
  cEnums : array[1..5] of String =
              (sTZFirst, sTZSecond, sTZThird, sTZFourth, sTZLast);
begin
  with ADate do begin
    if (wMonth in [1..12]) and (wDay in [1..5]) and (wYear=0) and
       (wDayOfWeek in [0..6]) then begin
      Result:=Format(sTZDateDesc,
        [cEnums[wDay],LongDayNames[wDayOfWeek+1],
         LongMonthNames[wMonth],
         FormatDateTime('tt',EncodeTime(wHour,wMinute,wSecond,wMilliseconds))]);
    end
    else begin
      Result:=sTZNotSet;
    end;
  end;
end;

function TimeZoneDateToDateTime(ADate:TSystemTime; AYear:Word):TDateTime;
var
  x : TDateTime;
begin
  with ADate do begin
    if (wMonth in [1..12]) and (wDay in [1..5]) and (wYear=0) and
       (wDayOfWeek in [0..6]) then begin
      x:=EncodeDate(AYear,wMonth,1);  // Monatsbeginn
      Result:=x+((wDayOfWeek-DayOfWeek(x)+8) mod 7)+7*(wDay-1);
      if wDay=5 then begin
        Inc(wMonth);
        if wMonth>12 then begin
          Dec(wMonth,12);
          Inc(AYear,1);
        end;
        if Result>=EncodeDate(AYear,wMonth,1) then Result:=Result-7;
        Result:=Result+EncodeTime(wHour,wMinute,wSecond,wMilliseconds);
      end;
    end
    else begin
      Result:=0;
    end;
  end;
end;

function TimeZoneDateToDateTimeStr(ADate:TSystemTime; AYear:Word):String;
var
  x : TDateTime;
begin
  x:=TimeZoneDateToDateTime(ADate,AYear);
  if x>0 then
    Result:=FormatDateTime('ddddd  tt',x)
  else
    Result:='';
end;

constructor TTimeZonesList.Create;
begin
  inherited Create;
  FList:=TStringList.Create;
  FCurrent:=-1;
  ReadList;
end;

destructor TTimeZonesList.Destroy;
begin
  ClearList;
  inherited Destroy;
end;

procedure TTimeZonesList.ClearList;
var
  i : Integer;
begin
  with FList do begin
    for i:=0 to Count-1 do begin
      FreeMem(Pointer(Objects[i]),SizeOf(TTimeZoneInformation));
    end;
    Clear;
  end;
end;

type
  TRegTZI = packed record
    tziBias         : LongInt;
    tziStandardBias : LongInt;
    tziDaylightBias : LongInt;
    tziStandardDate : TSystemTime;
    tziDaylightDate : TSystemTime;
  end;

procedure TTimeZonesList.ReadList;
const
  cRegPath = '\Software\Microsoft\Windows\CurrentVersion\Time Zones';
var
  l : TStringList;
  i : Integer;
  t : PTimeZoneInformation;
  tzi : TRegTZI;
  Display : String;
  CurrentTZ : String;
begin
  with TRegistry.Create do try
    ClearList;
    RootKey:=HKEY_LOCAL_MACHINE;
    if OpenKey(cRegPath,False) then begin
      l:=TStringList.Create;
      try
        GetKeyNames(l);
        for i:=0 to l.Count-1 do begin
          if OpenKey(Format('%s\%s',[cRegPath,l[i]]),False) then begin
            if Not ValueExists('Display') then Continue;
            GetMem(t,SizeOf(TTimeZoneInformation));
            try
              Display:=ReadString('Display');
              FillChar(t^,SizeOf(TTimeZoneInformation),0);
              with t^ do begin
                StringToWideChar(ReadString('Std'),StandardName,32);
                StringToWideChar(ReadString('Dlt'),DaylightName,32);
                ReadBinaryData('TZI',tzi,SizeOf(TZI));
                with tzi do begin
                  Bias        :=tziBias;
                  StandardDate:=tziStandardDate;
                  StandardBias:=tziStandardBias;
                  DaylightDate:=tziDaylightDate;
                  DaylightBias:=tziDaylightBias;
                end;
              end;
              FList.AddObject(Display,Pointer(t));
            except
              FreeMem(t,SizeOf(TTimeZoneInformation));
            end;
          end;
        end;
      finally
        l.Free;
      end;
      FList.Sort;
      // Suchen nach dem aktuellen Eintrag
      if OpenKey(cRegPath,False) then begin
        if ValueExists('') then begin
          CurrentTZ:=ReadString('');
          with FList do begin
            for i:=0 to Count-1 do begin
              with PTimeZoneInformation(Objects[i])^ do begin
                if AnsiCompareText(WideCharToString(StandardName),CurrentTZ)=0 then begin
                  FCurrent:=i;
                  Break;
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  finally
    Free;
  end;
end;

function TTimeZonesList.GetCount:Integer;
begin
  Result:=FList.Count;
end;

function TTimeZonesList.GetItems(Index:Integer):TTimeZoneInformation;
begin
  Result:=PTimeZoneInformation(FList.Objects[Index])^;
end;

function TTimeZonesList.IndexOf(const AName:String):Integer;
begin
  Result:=FList.IndexOf(AName);
end;

procedure TTimeZonesList.AssignList(AList:TStrings);
var
  i : Integer;
begin
  with AList do begin
    BeginUpdate;
    try
      Clear;
      for i:=0 to FList.Count-1 do begin
        Add(FList[i]);
      end;
    finally
      EndUpdate;
    end;
  end;
end;

// *******************************************************************************************

//  History:
//   2005-02-20, Peter J. Haas
//    - new license
//
//   2000-04-08, Peter J. Haas
//    - Version 1.0

end.
